from .airwave import AirWave
#from .clearpass import ClearPass, ClearPassDB
from .mmclient import MMClient

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
